# -*- coding: utf-8 -*-
import importlib
import logging

from django.db.models import Q
from rest_framework import exceptions

from ..utils import ApiRole

# TODO: logger console va bene nei pacchetti?
logger = logging.getLogger('console')
logger.setLevel(logging.INFO)

# TODO: non c'è nemmeno un'eccezione gestita, sistemare!


class ThuxRoleViewSetMixin:
    def write_log(self):
        logger.info('-----------------------------------')
        logger.info(' VIEWSET ')
        logger.info('-----------------------------------')
        logger.info(f'app label:        {self.app_label}')
        logger.info(f'model name:       {self.model_name}')
        logger.info(f'model class:      {self.model_class}')
        logger.info(f'action:           {self.action}')
        logger.info(f'has action:       {self.has_action}')
        logger.info(f'role codename:    {self.role_codename}')
        logger.info('-----------------------------------')
        logger.info(' SERIALIZER ')
        logger.info('-----------------------------------')
        logger.info(f'object:           {self.object}')
        logger.info(f'from status:      {self.from_status}')
        logger.info(f'to status:        {self.to_status}')
        logger.info(f'used fields:      {self.used_fields}')
        logger.info(f'required fields:  {self.required_fields}')
        logger.info(f'filters:          {self.filters}')
        logger.info('-----------------------------------')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.ar = None
        self.app_label = None
        self.model_name = None
        self.model_class = None
        self.action = None
        self.has_action = False
        self.role_codename = None
        self.object = None
        self.from_status = None
        self.to_status = []
        self.used_fields = []
        self.required_fields = []
        self.filters = None

    def _initial(self, request, *args, **kwargs):
        self.set_app_and_model_names()
        self.set_ar()
        self.set_role_codename(**kwargs)
        self.set_object(**kwargs)
        self.set_from_status()
        self.set_to_status()
        self.set_permission()
        self.set_filters()
        self.set_used_fields()
        self.set_required_fields()
        self.write_log()

    def initial(self, request, *args, **kwargs):
        self._initial(request, *args, **kwargs)
        super().initial(request, *args, **kwargs)

    def set_app_and_model_names(self):
        qs = self.queryset
        self.app_label = qs.model._meta.app_label
        self.model_name = qs.model.__name__.lower()
        self.model_class = qs.model.__name__

    def set_ar(self):
        self.ar = ApiRole(self.request.user, self.app_label, self.model_name)

    def set_role_codename(self, **kwargs):
        self.role_codename = kwargs.get('profile', None)

    def set_object(self, **kwargs):
        qs = self.queryset
        try:
            lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field
            filter_kwargs = {self.lookup_field: self.kwargs[lookup_url_kwarg]}
            self.object = qs.get(**filter_kwargs)
        except Exception:
            self.object = None

    def set_from_status(self):
        try:
            self.from_status = self.object.status
        except Exception:
            pass

    def set_to_status(self):
        self.to_status = self.ar.get_to_status(from_status=self.from_status, role_codename=self.role_codename)

    def set_permission(self):
        self.has_action = self.ar.has_action(self.action, role_codename=self.role_codename)

    def set_filters(self):
        self.filters = self.ar.get_filters(action=self.action, role_codename=self.role_codename)

    def set_required_fields(self):
        try:
            self.required_fields = self.ar.get_required_fields(
                from_status=self.from_status, role_codename=self.role_codename
            )
        except Exception:
            pass

    def set_used_fields(self):
        try:
            self.used_fields = self.ar.get_used_fields(
                from_status=self.from_status, action=self.action,
                role_codename=self.role_codename
            )
        except Exception:
            pass

    # GENERIC VIEWSET METHODS
    def get_queryset(self):
        qs = super().get_queryset()
        if self.request.user and self.request.user.is_superuser:
            return qs
        try:
            # TODO: make it safe!
            cond = eval(self.filters)
            qs = qs.filter(Q(cond))
            return qs
        except Exception:
            return qs.none()

    def check_permissions(self, request):
        super().check_permissions(request)
        if request.user and request.user.is_superuser:
            return
        if self.has_action:
            return
        raise exceptions.PermissionDenied()

    def get_serializer_class(self):
        action = self.action.replace('-', ' ').replace('_', ' ').title().replace(' ', '')
        model = self.model_class
        serializer = getattr(
            importlib.import_module(self.serializer_class.__module__), f'{model}{action}Serializer', None
        )
        if serializer:
            return serializer
        return super().get_serializer_class()
