# -*- coding: utf-8 -*-

from django.apps import apps
from django.apps.registry import Apps
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.utils.module_loading import import_string

from . import models

User = get_user_model()


class ApiRole:
    def __init__(self, user, app_label, model_name, *args, **kwargs):
        """
        init method
        :param user: user instance
        :param app_label: app label Django
        :param model_name: model name Django
        :param args: not used
        :param kwargs: not used
        """
        self.user = user if not user.is_anonymous else None
        self.app_label = app_label
        self.model_name = model_name

    def get_all_roles(self):
        return models.Role.objects.filter(
            profile__users=self.user,
            status=1,
        ).distinct()

    def get_all_profiles(self):
        return models.Profile.objects.filter(
            users=self.user,
            status=1,
        ).values_list('codename', flat=True).order_by('codename').distinct('codename')

    def get_app_profiles(self):
        return models.Profile.objects.filter(
            users=self.user,
            roles__app_label=self.app_label,
            status=1,
        ).values('codename').order_by('codename').distinct('codename')

    def get_roles(self):
        """
        Return user roles for app and model assigned
        :return: queryset
        """
        return models.Role.objects.filter(
            app_label=self.app_label,
            model_name=self.model_name,
            profile__users=self.user,
            status=1,
        ).distinct()

    def get_role_codenames(self):
        """
        Return user role codenames for app and model assigned
        :return: list
        """
        # TODO: deprecation warning
        return self.get_profile_codenames()

    def get_profile_codenames(self):
        """
        Return user profile codenames for app and model assigned
        :return: list
        """
        return list(self.get_roles().values_list('profile__codename', flat=True).distinct())

    def get_filters(self, action, role_codename=None):
        """
        Return user filters for app, model and action assigned
        :param action: action name defined in DRF views
        :param role_codename: codename field of Role model
        :return: string
        """
        objs = models.Filter.objects.filter(
            status=1,
            role__status=1,
            role__app_label=self.app_label,
            role__model_name=self.model_name,
            role__profile__users=self.user,
            codename=action,
        )
        if role_codename:
            objs = objs.filter(role__profile__codename=role_codename).distinct()
        obj_q = []
        if objs:
            for obj in objs:
                obj_q.append(obj.code)
        return "|".join(obj_q)

    def has_action(self, action, role_codename=None):
        """
        Return True or False if user has action passed
        :param action: action name defined in DRF views
        :param role_codename: codename field of Role model
        :return: Boolean
        """
        objs = models.Filter.objects.filter(
            status=1,
            role__status=1,
            role__app_label=self.app_label,
            role__model_name=self.model_name,
            role__profile__users=self.user,
            codename=action,
        )
        if role_codename:
            objs = objs.filter(role__profile__codename=role_codename)
        return objs.exists()

    def get_required_fields(self, from_status, role_codename=None):
        """
        Return user required fields for app, model and from_status assigned
        :param from_status: object status
        :param role_codename: codename field of Role model
        :return: list
        """
        from_status = str(from_status)
        objs = models.RequiredField.objects.filter(
            status=1,
            role__status=1,
            role__app_label=self.app_label,
            role__model_name=self.model_name,
            role__profile__users=self.user,
            from_status=from_status,
        )
        if role_codename:
            objs = objs.filter(role__profile__codename=role_codename).distinct()
        fields = []
        if objs:
            for obj in objs:
                if obj.fields == '__all__':
                    return obj.fields
        fields += obj.fields
        return list(set(fields))

    def get_used_fields(self, from_status, action, role_codename=None):
        """
        Return user used fields for app, model, from status and action assigned
        :param from_status: object status
        :param action: action name defined in DRF views
        :param role_codename: codename field of Role model
        :return: list
        """
        objs = models.UsedField.objects.filter(
            status=1,
            role__status=1,
            role__app_label=self.app_label,
            role__model_name=self.model_name,
            role__profile__users=self.user,
            action=action,
        )
        if from_status is not None:
            objs = objs.filter(from_status=from_status)
        if role_codename:
            objs = objs.filter(role__profile__codename=role_codename).distinct()
        fields = []
        if objs:
            for obj in objs:
                if obj.fields == '__all__':
                    return obj.fields
        fields += obj.fields
        return list(set(fields))

    def get_to_status(self, from_status, role_codename=None):
        """
        Return list of user "to status" for app, model and from status assigned
        :param from_status: object status
        :param role_codename: codename field of Role model
        :return: list
        """
        objs = models.StatusFlow.objects.filter(
            status=1,
            role__status=1,
            role__app_label=self.app_label,
            role__model_name=self.model_name,
            role__profile__users=self.user,
            from_status=from_status,
        )
        if role_codename:
            objs = objs.filter(role__profile__codename=role_codename).distinct()
        to_status = []
        if objs:
            for obj in objs:
                to_status += obj.to_status
        return list(set(to_status))


def get_role_settings(**kwargs):
    """Return prepared role settings"""
    from django.conf import settings

    # default
    app_settings = settings.THX_DEFAULT_ROLE.copy()
    # project settings
    app_settings.update(getattr(settings, 'THX_ROLE', {}))
    # local usage override
    app_settings.update(kwargs)

    # load and run functions if necessary
    for k, v in app_settings.items():
        if k.endswith("_FUN") and v:
            app_settings[k.replace("_FUN", "")] = import_string(v)()

    return app_settings


def get_role_apps(role_apps):
    """Return apps defined to be initialized with thx.role package."""
    return apps if role_apps == "__all__" else [apps.get_app_config(app_label) for app_label in role_apps]


class InitialRole:
    def __init__(self, **kwargs):
        self.user = User.objects.get(pk=1)

        app_settings = get_role_settings(**kwargs)
        self.apps = get_role_apps(app_settings.get('APPS', "__all__"))
        self.models = self.get_models(app_settings.get('MODELS', "__all__"))
        self.roles = app_settings["ROLES"]
        self.status = app_settings["STATUS"]
        self.actions = app_settings["ACTIONS"]
        self.actions_no_fields = app_settings["ACTIONS_NO_FIELDS"]
        self.used_fields = app_settings["USED_FIELDS"]
        self.req_fields = app_settings["REQ_FIELDS"]

    def get_models(self, role_models):
        """
        Example: ("carriage.Carriage", "carriage.Delivery")
        """
        if role_models == "__all__":
            if isinstance(self.apps, Apps):
                role_models = [model._meta.label for model in self.apps.get_models()]
            else:
                # FixMe se inizializzo self.apps come Apps(installed_apps=self.apps),
                #  self.apps.get_models() ritorna una lista vuota.
                role_models = [model._meta.label for app in self.apps for model in app.get_models()]
        return role_models

    def insert_all(self, app_name=None, model_name=None):
        self.insert_profiles()
        self.insert_roles()
        self.insert_used_fields(app_name, model_name)
        self.insert_required_fields(app_name, model_name)
        self.insert_status_flows(app_name, model_name)
        self.insert_filters(app_name, model_name)

    def insert_all_app(self, app_name):
        self.models = [model._meta.label for model in apps.get_app_config(app_name).get_models()]
        self.insert_all(app_name)

    def insert_all_model(self, app_name, model_name):
        self.models = [f'{app_name}.{model_name}']
        self.insert_all(app_name, model_name)

    def insert_profiles(self):
        for profile in self.roles:
            profile = profile.lower()
            try:
                models.Profile.objects.create(
                    creator=self.user, last_modifier=self.user,
                    name=profile,
                )
            except ValidationError:
                pass

    def insert_roles(self):
        for model in self.models:
            app_label, model_name = model.split('.')
            app_label = app_label.lower()
            model_name = model_name.title()
            try:
                models.Role.objects.create(
                    creator=self.user, last_modifier=self.user,
                    app_label=app_label, model_name=model_name,
                )
            except ValidationError:
                pass

    def insert_used_fields(self, app_name=None, model_name=None):
        role_filters = {
            'status': 1
        }
        if app_name:
            role_filters['app_label'] = app_name
        if model_name:
            role_filters['model_name'] = model_name

        roles = models.Role.objects.filter(**role_filters)
        for role in roles:
            for status in self.status:
                for action in self.actions:
                    try:
                        models.UsedField.objects.create(
                            creator=self.user, last_modifier=self.user,
                            role=role, from_status=status, action=action, fields=self.used_fields,
                        )
                    except ValidationError:
                        pass

    def insert_required_fields(self, app_name=None, model_name=None):
        role_filters = {
            'status': 1
        }
        if app_name:
            role_filters['app_label'] = app_name
        if model_name:
            role_filters['model_name'] = model_name
        roles = models.Role.objects.filter(**role_filters)
        for role in roles:
            for status in self.status:
                try:
                    models.RequiredField.objects.create(
                        creator=self.user, last_modifier=self.user,
                        role=role, from_status=status, fields=self.req_fields,
                    )
                except ValidationError:
                    pass

    def insert_status_flows(self, app_name=None, model_name=None):
        role_filters = {
            'status': 1
        }
        if app_name:
            role_filters['app_label'] = app_name
        if model_name:
            role_filters['model_name'] = model_name
        roles = models.Role.objects.filter(**role_filters)
        for role in roles:
            for from_status in self.status:
                for to_status in self.status:
                    if from_status != to_status:
                        try:
                            if not isinstance(to_status, list):
                                to_status = [to_status]
                            models.StatusFlow.objects.create(
                                creator=self.user, last_modifier=self.user,
                                role=role, from_status=from_status, to_status=to_status,
                            )
                        except ValidationError:
                            pass

    def insert_filters(self, app_name=None, model_name=None):
        role_filters = {
            'status': 1
        }
        if app_name:
            role_filters['app_label'] = app_name
        if model_name:
            role_filters['model_name'] = model_name
        roles = models.Role.objects.filter(**role_filters)
        for role in roles:
            for action in self.actions:
                try:
                    models.Filter.objects.create(
                        creator=self.user, last_modifier=self.user,
                        role=role, action=action,
                    )
                except ValidationError:
                    pass
            for action in self.actions_no_fields:
                try:
                    models.Filter.objects.create(
                        creator=self.user, last_modifier=self.user,
                        role=role, action=action,
                    )
                except ValidationError:
                    pass
